@@ -2,6 +2,7 @@
 Master Control Program
 
 """
+import collections
 import logging
 import multiprocessing
 import os
@@ -14,12 +15,13 @@ import signal
 import sys
 import time
 
-from rejected import state
-from rejected import process
-from rejected import __version__
+from rejected import state, process, __version__
 
 LOGGER = logging.getLogger(__name__)
 
+_PROCESS_RUNNING = [psutil.STATUS_RUNNING, psutil.STATUS_SLEEPING]
+_PROCESS_STOPPED_OR_DEAD = [psutil.STATUS_STOPPED, psutil.STATUS_DEAD]
+
 
 class Consumer(object):
     """Class used for keeping track of each consumer type being managed by
@@ -27,8 +29,7 @@ class Consumer(object):
 
     """
 
-    def __init__(self, connections, last_proc_num, processes, qty, queue):
-        self.connections = connections
+    def __init__(self, last_proc_num, processes, qty, queue):
         self.last_proc_num = last_proc_num
         self.processes = processes
         self.qty = qty
@@ -40,6 +41,7 @@ class MasterControlProgram(state.State):
 
     DEFAULT_CONSUMER_QTY = 1
     MAX_SHUTDOWN_WAIT = 10
+    MAX_UNRESPONSIVE_COUNT = 3
     POLL_INTERVAL = 60.0
     POLL_RESULTS_INTERVAL = 3.0
     SHUTDOWN_WAIT = 1
@@ -59,17 +61,19 @@ class MasterControlProgram(state.State):
         super(MasterControlProgram, self).__init__()
 
         # Default values
+        self._active_cache = None
         self.consumer_cfg = self.get_consumer_cfg(config, consumer, quantity)
         self.consumers = dict()
         self.config = config
         self.last_poll_results = dict()
-        self.poll_data = {'time': 0, 'processes': list()}
+        self.poll_data = {'time': 0, 'processes': []}
         self.poll_timer = None
         self.profile = profile
         self.results_timer = None
         self.stats = dict()
         self.stats_queue = multiprocessing.Queue()
         self.polled = False
+        self.unresponsive = collections.Counter()
 
         # Flag to indicate child creation error
         self.child_abort = False
@@ -84,18 +88,24 @@ class MasterControlProgram(state.State):
                                                     self.POLL_INTERVAL)
         LOGGER.debug('Set process poll interval to %.2f', self.poll_interval)
 
-    @property
-    def active_processes(self):
+    def active_processes(self, use_cache=True):
         """Return a list of all active processes, pruning dead ones
 
         :rtype: list
 
         """
+        LOGGER.debug('Checking active processes (cache: %s)', use_cache)
+        if use_cache and self._active_cache and \
+                self._active_cache[0] > time.time() - self.poll_interval:
+            return self._active_cache[1]
         active_processes, dead_processes = list(), list()
         for consumer in self.consumers:
             for name in self.consumers[consumer].processes:
                 child = self.get_consumer_process(consumer, name)
-                if int(child.pid) == os.getpid():
+                if child.pid is None:
+                    dead_processes.append((consumer, name))
+                    continue
+                elif int(child.pid) == os.getpid():
                     continue
                 try:
                     proc = psutil.Process(child.pid)
@@ -103,7 +113,16 @@ class MasterControlProgram(state.State):
                     dead_processes.append((consumer, name))
                     continue
 
-                if self.is_dead(proc):
+                if self.unresponsive[name] >= self.MAX_UNRESPONSIVE_COUNT:
+                    LOGGER.info('Killing unresponsive consumer %s (%i): '
+                                '%i misses',
+                                name, proc.pid, self.unresponsive[name])
+                    try:
+                        os.kill(child.pid, signal.SIGABRT)
+                    except OSError:
+                        pass
+                    dead_processes.append((consumer, name))
+                elif self.is_dead(proc, name):
                     dead_processes.append((consumer, name))
                 else:
                     active_processes.append(child)
@@ -112,6 +131,7 @@ class MasterControlProgram(state.State):
             LOGGER.debug('Removing %i dead process(es)', len(dead_processes))
             for proc in dead_processes:
                 self.remove_consumer_process(*proc)
+        self._active_cache = time.time(), active_processes
         return active_processes
 
     def calculate_stats(self, data):
@@ -123,15 +143,13 @@ class MasterControlProgram(state.State):
         """
         timestamp = data['timestamp']
         del data['timestamp']
-        LOGGER.debug('Calculating stats for data timestamp: %i', timestamp)
 
         # Iterate through the last poll results
         stats = self.consumer_stats_counter()
         consumer_stats = dict()
         for name in data.keys():
             consumer_stats[name] = self.consumer_stats_counter()
-            consumer_stats[name]['processes'] = \
-                self.process_count_by_consumer(name)
+            consumer_stats[name]['processes'] = self.process_count(name)
             for proc in data[name].keys():
                 for key in stats:
                     value = data[name][proc]['counts'].get(key, 0)
@@ -139,7 +157,7 @@ class MasterControlProgram(state.State):
                     consumer_stats[name][key] += value
 
         # Return a data structure that can be used in reporting out the stats
-        stats['processes'] = len(self.active_processes)
+        stats['processes'] = len(self.active_processes())
         return {
             'last_poll': timestamp,
             'consumers': consumer_stats,
@@ -154,12 +172,11 @@ class MasterControlProgram(state.State):
         """
         LOGGER.debug('Checking minimum consumer process levels')
         for name in self.consumers:
-            for connection in self.consumers[name].connections:
-                processes_needed = self.process_spawn_qty(name, connection)
-                LOGGER.debug('Need to spawn %i processes for %s on %s',
-                             processes_needed, name, connection)
-                if processes_needed:
-                    self.start_processes(name, connection, processes_needed)
+            processes_needed = self.process_spawn_qty(name)
+            if processes_needed:
+                LOGGER.info('Need to spawn %i processes for %s',
+                            processes_needed, name)
+                self.start_processes(name, processes_needed)
 
     def collect_results(self, data_values):
         """Receive the data from the consumers polled and process it.
@@ -231,33 +248,44 @@ class MasterControlProgram(state.State):
         """
         consumers = dict(config.application.Consumers)
         if only:
-            for key in consumers.keys():
+            for key in list(consumers.keys()):
                 if key != only:
                     del consumers[key]
             if qty:
                 consumers[only]['qty'] = qty
         return consumers
 
-    @staticmethod
-    def is_dead(proc):
+    def is_dead(self, proc, name):
         """Checks to see if the specified process is dead.
 
         :param psutil.Process proc: The process to check
+        :param str name: The name of consumer
         :rtype: bool
 
         """
-        status = proc.status()
-        LOGGER.debug('Process status (%r): %r', proc.pid, status)
+        LOGGER.debug('Checking %s (%r)', name, proc)
         try:
-            if status == psutil.STATUS_ZOMBIE:
-                proc.terminate()
-                return True
-            elif status == psutil.STATUS_DEAD:
-                return True
+            status = proc.status()
         except psutil.NoSuchProcess:
-            LOGGER.debug('Process is dead and does not exist')
+            LOGGER.debug('NoSuchProcess: %s (%r)', name, proc)
             return True
-        return False
+
+        LOGGER.debug('Process %s (%s) status: %r (Unresponsive Count: %s)',
+                     name, proc.pid, status, self.unresponsive[name])
+        if status in _PROCESS_RUNNING:
+            return False
+        elif status == psutil.STATUS_ZOMBIE:
+            try:
+                proc.wait(0.1)
+            except psutil.TimeoutExpired:
+                pass
+            try:
+                proc.terminate()
+                status = proc.status()
+            except psutil.NoSuchProcess:
+                LOGGER.debug('NoSuchProcess: %s (%r)', name, proc)
+                return True
+        return status in _PROCESS_STOPPED_OR_DEAD
 
     def kill_processes(self):
         """Gets called on shutdown by the timer when too much time has gone by,
@@ -266,17 +294,20 @@ class MasterControlProgram(state.State):
 
         """
         LOGGER.critical('Max shutdown exceeded, forcibly exiting')
-        processes = True
+        processes = self.active_processes(False)
         while processes:
-            processes = self.active_processes
-            for proc in processes:
+            for proc in self.active_processes(False):
                 if int(proc.pid) != int(os.getpid()):
                     LOGGER.warning('Killing %s (%s)', proc.name, proc.pid)
-                    os.kill(int(proc.pid), signal.SIGKILL)
+                    try:
+                        os.kill(int(proc.pid), signal.SIGKILL)
+                    except OSError:
+                        pass
                 else:
                     LOGGER.warning('Cowardly refusing kill self (%s, %s)',
                                    proc.pid, os.getpid())
             time.sleep(0.5)
+            processes = self.active_processes(False)
 
         LOGGER.info('Killed all children')
         return self.set_state(self.STATE_STOPPED)
@@ -305,32 +336,30 @@ class MasterControlProgram(state.State):
                         self.stats['consumers'][key]['processed'],
                         self.stats['consumers'][key]['failed'])
 
-    def new_consumer(self, config):
+    def new_consumer(self, config, consumer_name):
         """Return a consumer dict for the given name and configuration.
 
         :param dict config: The consumer configuration
+        :param str consumer_name: The consumer name
         :rtype: dict
 
         """
-        return Consumer(dict([(c, []) for c in config['connections']]), 0,
-                        dict(), config.get('qty', self.DEFAULT_CONSUMER_QTY),
-                        config['queue'])
+        return Consumer(0,
+                        dict(),
+                        config.get('qty', self.DEFAULT_CONSUMER_QTY),
+                        config.get('queue', consumer_name))
 
-    def new_process(self, consumer_name, connection_name):
+    def new_process(self, consumer_name):
         """Create a new consumer instances
 
         :param str consumer_name: The name of the consumer
-        :param str connection_name: The name of the connection
         :return tuple: (str, process.Process)
 
         """
         process_name = '%s-%s' % (consumer_name,
                                   self.new_process_number(consumer_name))
-        LOGGER.debug('Creating a new process for %s: %s', connection_name,
-                     process_name)
         kwargs = {
             'config': self.config.application,
-            'connection_name': connection_name,
             'consumer_name': consumer_name,
             'profile': self.profile,
             'daemon': False,
@@ -350,14 +379,26 @@ class MasterControlProgram(state.State):
         self.consumers[name].last_proc_num += 1
         return self.consumers[name].last_proc_num
 
-    def on_abort(self, _signum, _unused_frame):
-        LOGGER.debug('Abort signal received from child')
-        time.sleep(1)
-        if not self.active_processes:
+    def on_sigchld(self, _signum, _unused_frame):
+        """Invoked when a child sends up an SIGCHLD signal.
+
+        :param int _signum: The signal that was invoked
+        :param frame _unused_frame: The frame that was interrupted
+
+        """
+        LOGGER.info('SIGCHLD received from child')
+        if not self.active_processes(False):
             LOGGER.info('Stopping with no active processes and child error')
+            signal.setitimer(signal.ITIMER_REAL, 0, 0)
             self.set_state(self.STATE_STOPPED)
 
     def on_timer(self, _signum, _unused_frame):
+        """Invoked by the Poll timer signal.
+
+        :param int _signum: The signal that was invoked
+        :param frame _unused_frame: The frame that was interrupted
+
+        """
         if self.is_shutting_down:
             LOGGER.debug('Polling timer fired while shutting down')
             return
@@ -374,6 +415,15 @@ class MasterControlProgram(state.State):
             if self.log_stats_enabled:
                 self.log_stats()
 
+            # Increment the unresponsive children
+            for proc_name in self.poll_data['processes']:
+                self.unresponsive[proc_name] += 1
+
+            # Remove counters for processes that came back to life
+            for proc_name in list(self.unresponsive.keys()):
+                if proc_name not in self.poll_data['processes']:
+                    del self.unresponsive[proc_name]
+
     def poll(self):
         """Start the poll process by invoking the get_stats method of the
         consumers. If we hit this after another interval without fully
@@ -391,10 +441,8 @@ class MasterControlProgram(state.State):
         self.poll_data = {'timestamp': time.time(), 'processes': list()}
 
         # Iterate through all of the consumers
-        for proc in self.active_processes:
-            LOGGER.debug('Checking runtime state of %s', proc.name)
+        for proc in list(self.active_processes()):
             if proc == multiprocessing.current_process():
-                LOGGER.debug('Matched current process in active_processes')
                 continue
 
             # Send the profile signal
@@ -410,7 +458,8 @@ class MasterControlProgram(state.State):
         :rtype: bool
 
         """
-        return (time.time() - self.poll_data['timestamp']) >= self.poll_interval
+        return ((time.time() - self.poll_data['timestamp']) >=
+                self.poll_interval)
 
     def poll_results_check(self):
         """Check the polling results by checking to see if the stats queue is
@@ -445,38 +494,23 @@ class MasterControlProgram(state.State):
         """
         return self.consumers[name].processes[process_name]
 
-    def process_count(self, name, connection):
-        """Return the process count for the given consumer name and connection.
-
-        :param str name: The consumer name
-        :param str connection: The connection name
-        :rtype: int
-
-        """
-        return len(self.consumers[name].connections[connection])
-
-    def process_count_by_consumer(self, name):
-        """Return the process count by consumer only.
+    def process_count(self, name):
+        """Return the process count for the given consumer name.
 
         :param str name: The consumer name
         :rtype: int
 
         """
-        count = 0
-        for connection in self.consumers[name].connections:
-            count += len(self.consumers[name].connections.get(connection))
-        return count
+        return len(self.consumers[name].processes)
 
-    def process_spawn_qty(self, name, connection):
-        """Return the number of processes to spawn for the given consumer name
-        and connection.
+    def process_spawn_qty(self, name):
+        """Return the number of processes to spawn for the given consumer name.
 
         :param str name: The consumer name
-        :param str connection: The connection name
         :rtype: int
 
         """
-        return self.consumers[name].qty - self.process_count(name, connection)
+        return self.consumers[name].qty - self.process_count(name)
 
     def remove_consumer_process(self, consumer, name):
         """Remove all details for the specified consumer and process name.
@@ -486,20 +520,25 @@ class MasterControlProgram(state.State):
 
         """
         my_pid = os.getpid()
-        for conn in self.consumers[consumer].connections:
-            if name in self.consumers[consumer].connections[conn]:
-                self.consumers[consumer].connections[conn].remove(name)
-        if name in self.consumers[consumer].processes:
+        if name in self.consumers[consumer].processes.keys():
             child = self.consumers[consumer].processes[name]
-            if child.is_alive():
-                if child.pid != my_pid:
+            try:
+                alive = child.is_alive()
+            except AssertionError:
+                LOGGER.debug('Tried to test non-child process (%r to %r)',
+                             os.getpid(), child.pid)
+            else:
+                if child.pid == my_pid:
+                    LOGGER.debug('Child has my pid? %r, %r', my_pid, child.pid)
+                elif alive:
                     try:
                         child.terminate()
                     except OSError:
                         pass
-                else:
-                    LOGGER.debug('Child has my pid? %r, %r', my_pid, child.pid)
-            del self.consumers[consumer].processes[name]
+            try:
+                del self.consumers[consumer].processes[name]
+            except KeyError:
+                pass
 
     def run(self):
         """When the consumer is ready to start running, kick off all of our
@@ -509,8 +548,8 @@ class MasterControlProgram(state.State):
         self.set_state(self.STATE_ACTIVE)
         self.setup_consumers()
 
-        # Set the SIGABRT handler for child creation errors
-        signal.signal(signal.SIGABRT, self.on_abort)
+        # Set the SIGCHLD handler for child creation errors
+        signal.signal(signal.SIGCHLD, self.on_sigchld)
 
         # Set the SIGALRM handler for poll interval
         signal.signal(signal.SIGALRM, self.on_timer)
@@ -554,44 +593,45 @@ class MasterControlProgram(state.State):
         # Set the signal timer
         signal.setitimer(signal.ITIMER_REAL, duration, 0)
 
-    def start_process(self, name, connection):
-        """Start a new consumer process for the given consumer & connection name
-
-        :param str name: The consumer name
-        :param str connection: The connection name
-
-        """
-        process_name, proc = self.new_process(name, connection)
-        LOGGER.info('Spawning %s process for %s to %s', process_name, name,
-                    connection)
-
-        # Append the process to the consumer process list
-        self.consumers[name].processes[process_name] = proc
-        self.consumers[name].connections[connection].append(process_name)
-
-        # Start the process
-        proc.start()
-
-    def start_processes(self, name, connection, quantity):
-        """Start the specified quantity of consumer processes for the given
-        consumer and connection.
-
-        :param str name: The consumer name
-        :param str connection: The connection name
-        :param int quantity: The quantity of processes to start
-
-        """
-        [self.start_process(name, connection) for _i in range(0, quantity)]
-
     def setup_consumers(self):
         """Iterate through each consumer in the configuration and kick off the
         minimal amount of processes, setting up the runtime data as well.
 
         """
         for name in self.consumer_cfg.keys():
-            self.consumers[name] = self.new_consumer(self.consumer_cfg[name])
-            for connection in self.consumers[name].connections:
-                self.start_processes(name, connection, self.consumers[name].qty)
+            self.consumers[name] = self.new_consumer(
+                self.consumer_cfg[name], name)
+            self.start_processes(name, self.consumers[name].qty)
+
+    def start_process(self, name):
+        """Start a new consumer process for the given consumer name
+
+        :param str name: The consumer name
+
+        """
+        process_name, proc = self.new_process(name)
+        LOGGER.info('Spawning %s process for %s', process_name, name)
+
+        # Append the process to the consumer process list
+        self.consumers[name].processes[process_name] = proc
+
+        # Start the process
+        try:
+            proc.start()
+        except IOError as error:
+            LOGGER.critical('Failed to start %s for %s: %r',
+                            process_name, name, error)
+            del self.consumers[name].process[process_name]
+
+    def start_processes(self, name, quantity):
+        """Start the specified quantity of consumer processes for the given
+        consumer.
+
+        :param str name: The consumer name
+        :param int quantity: The quantity of processes to start
+
+        """
+        [self.start_process(name) for i in range(0, quantity or 0)]
 
     def stop_processes(self):
         """Iterate through all of the consumer processes shutting them down."""
@@ -608,27 +648,26 @@ class MasterControlProgram(state.State):
         LOGGER.info('Sending SIGABRT to active children')
         for proc in multiprocessing.active_children():
             if int(proc.pid) != os.getpid():
-                os.kill(int(proc.pid), signal.SIGABRT)
+                try:
+                    os.kill(int(proc.pid), signal.SIGABRT)
+                except OSError:
+                    pass
 
         # Wait for them to finish up to MAX_SHUTDOWN_WAIT
-        iterations = 0
-        processes = self.total_process_count
-        while processes:
-            LOGGER.info('Waiting on %i active processes to shut down',
-                        processes)
+        for iteration in range(0, self.MAX_SHUTDOWN_WAIT):
+            processes = len(self.active_processes(False))
+            if not processes:
+                break
+
+            LOGGER.info('Waiting on %i active processes to shut down (%i/%i)',
+                        processes, iteration, self.MAX_SHUTDOWN_WAIT)
             try:
                 time.sleep(0.5)
             except KeyboardInterrupt:
-                LOGGER.info('Caught CTRL-C, Killing Children')
-                self.kill_processes()
-                self.set_state(self.STATE_STOPPED)
-                return
-
-            iterations += 1
-            if iterations == self.MAX_SHUTDOWN_WAIT:
-                self.kill_processes()
                 break
-            processes = self.total_process_count
+
+        if len(self.active_processes(False)):
+            self.kill_processes()
 
         LOGGER.debug('All consumer processes stopped')
         self.set_state(self.STATE_STOPPED)
@@ -640,4 +679,4 @@ class MasterControlProgram(state.State):
         :rtype: int
 
         """
-        return len(self.active_processes)
+        return len(self.active_processes(False))
